private class SudokuCellView : Gtk.DrawingArea
{
    private Pango.Layout layout;

    private SudokuCell _cell;
    public SudokuCell cell
    {
        set
        {
            _cell = value;
            this.value = cell.value;
            _cell.value_changed.connect (value_changed_cb);
        }
        get { return _cell; }
    }

    private string _text;
    public string text
    {
        set
        {
            _text = value;
            layout = create_pango_layout (_text);
            layout.set_font_description (style.font_desc);
            queue_draw ();
        }
        get { return _text; }
    }
    
    public int value
    {
        set
        {
            if (value == 0)
                text = "";
            else
                text = "%d".printf (value);
        }
    }

    public SudokuCellView ()
    {
        can_focus = true;
        style.font_desc.set_size (Pango.SCALE * 13);
        text = "";
    }

    public override void size_request (out Gtk.Requisition requisition)
    {
        int width, height, side;
        layout.get_size (out width, out height);
        side = width > height ? width : height;
        requisition.width = requisition.height = side / Pango.SCALE;
    }
    
    public override bool key_release_event (Gdk.EventKey event)
    {
        // FIXME: Can't find vala bindings to Gdk keyvals...

        if (event.keyval >= 49 /* 1 */ && event.keyval <= 58 /* 9 */)
        {
            cell.value = (int)event.keyval - 48;
            return true;
        }

        if (event.keyval == 48 || event.keyval == 65288 /* backspace */ || event.keyval == 65535 /* delete */)
        {
            cell.value = 0;
            return true;
        }

        return false;
    }

    public override bool expose_event (Gdk.EventExpose event)
    {
        Cairo.Context c = Gdk.cairo_create (window);

        if (is_focus)
            c.set_source_rgb (1.0, 0.0, 0.0);
        else
            c.set_source_rgb (1.0, 1.0, 1.0);
        c.rectangle (0, 0, allocation.width, allocation.height);
        c.fill ();

        int glyph_width, glyph_height;
        layout.get_pixel_size (out glyph_width, out glyph_height);
        c.set_source_rgb (0.0, 0.0, 0.0);
        c.move_to ((allocation.width - glyph_width) / 2, (allocation.height - glyph_height) / 2);
        Pango.cairo_update_layout (c, layout);
        Pango.cairo_show_layout (c, layout);

        return false;
    }

    private void value_changed_cb (SudokuCell cell)
    {
        value = cell.value;
    }
}

public class SudokuView : Gtk.AspectFrame
{
    private SudokuGame game;
    private SudokuCellView[,] cells;

    public SudokuView (SudokuGame game)
    {
        this.game = game;

        shadow_type = Gtk.ShadowType.NONE;
        
        /* Use an EventBox to be able to set background */
        var box = new Gtk.EventBox ();
        box.modify_bg (Gtk.StateType.NORMAL, box.style.black);
        add (box);
        box.show ();

        var table = new Gtk.Table (9, 9, true);
        table.row_spacing = 1;
        table.column_spacing = 1;
        table.set_row_spacing (2, 2);
        table.set_row_spacing (5, 2);
        table.set_col_spacing (2, 2);
        table.set_col_spacing (5, 2);
        table.border_width = 2;

        cells = new SudokuCellView[9,9];
        for (int row = 0; row < 9; row++)
            for (int col = 0; col < 9; col++)
            {
                var cell = new SudokuCellView ();
                cells[row, col] = cell;
                cell.cell = game.cells[row, col];
                cell.show ();
                table.attach_defaults (cell, col, col+1, row, row+1);
            }
        box.add (table);
        table.show ();
    }
}
